/*
 * Copyright (c) 2008 - 2012, Andy Bierman, All Rights Reserved.
 * Copyright (c) 2012 - 2020, YumaWorks, Inc., All Rights Reserved.
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *


    Combined SIL module
    module sethook-example
    revision 2020-12-14
    namespace http://netconfcentral.org/ns/sethook-example
    Created: 2020-04-12T20:52:14Z
    CLI parameters:
        format c
        indent 4
        module sethook-example
        output sethook-example.c
        sil-edit2
        sil-get2
        unified true

 */

#include <xmlstring.h>

#include "procdefs.h"
#include "agt.h"
#include "agt_action.h"
#include "agt_cb.h"
#include "agt_cfg.h"
#include "agt_rpc.h"
#include "agt_sil_lib.h"
#include "agt_timer.h"
#include "agt_util.h"
#include "agt_val.h"
#include "cfg.h"
#include "dlq.h"
#include "getcb.h"
#include "ncx.h"
#include "ncx_feature.h"
#include "ncxmod.h"
#include "ncxtypes.h"
#include "obj.h"
#include "rpc.h"
#include "status.h"
#include "val.h"
#include "val_child.h"
#include "val_util.h"
#include "xml_val.h"
#include "sethook-example.h"

/* module static variables */
static ncx_module_t *sethook_example_mod;
static obj_template_t *interfaces_obj;
static obj_template_t *trigger_obj;
static val_value_t *interfaces_val;
static val_value_t *trigger_val;

/* put your static variables here */


/********************************************************************
* FUNCTION print_callback_info
*
* Dump current Callback info
*
********************************************************************/
static void
    print_callback_info (val_value_t *useval,
                         agt_cbtyp_t cbtyp,
                         op_editop_t editop,
                         const xmlChar *callback_name)
{
    if (LOGDEBUG) {
        log_debug("\n (%s) Enter %s callback"
            "\n ------PHASE:%s;EDITOP:%s\n",
            useval ? VAL_NAME(useval) : NCX_NT_NONE,
            callback_name,
            agt_cbtype_name(cbtyp),
            op_editop_name(editop));
    }

} /* print_callback_info */


/********************************************************************
* FUNCTION print_children
*
* Dump children values
*
********************************************************************/
static void print_children (val_value_t *val,
                            const char *which)
{
    val_value_t *child;

    log_debug("\n    %s %s P:[%p] :",
              which,
              val ? VAL_NAME(val) : NCX_EL_NULL,
              val);
    if (val) {
        val_dump_value(val, 10, DBG4);

        child = val_get_first_child(val);

        while (child) {

            if (child->btyp && typ_is_simple(child->btyp)) {
                xmlChar *valstr = val_make_sprintf_string(child);
                if (valstr == NULL) {
                    SET_ERROR(ERR_INTERNAL_MEM);
                    return;
                }

                log_debug("\nchild Q of simple childs:   %s='%s'",
                          VAL_NAME(child),
                          valstr);

                m__free(valstr);
            }

            child = val_get_next_child(child);
        }
    }
} /* print_children */


/********************************************************************
* FUNCTION handle_merge
*
* Dump children values during MERGE operation
*
********************************************************************/
static void handle_merge (rpc_msg_t *msg,
                          val_value_t *newval,
                          val_value_t *curval)
{
    print_children(newval, "newval");
    print_children(curval, "curval");

    log_debug("\n    Merge list:");

    /* the edit is not really on this node; need to get
     * each child_undo record to get the real edited nodes */
    agt_cfg_transaction_t *txcb = RPC_MSG_TXCB(msg);
    agt_cfg_undo_rec_t *child_edit =
        agt_cfg_first_child_edit(txcb, newval, curval);

    while (child_edit) {

        log_debug("\nchild_edit = %p", child_edit);
        log_debug("\nchild_edit->newnode = P:[%p]", child_edit->newnode);
        log_debug("\nchild_edit->curnode = P:[%p]", child_edit->curnode);

        op_editop_t child_editop = OP_EDITOP_NONE;
        val_value_t *child_newval = NULL;
        val_value_t *child_curval = NULL;
        xmlChar *newval_str = NULL;
        xmlChar *curval_str = NULL;

        agt_cfg_child_edit_fields(child_edit,
            &child_editop,
            &child_newval,
            &child_curval);

        if (child_newval) {
            newval_str = val_make_sprintf_string(child_newval);
            if (newval_str == NULL) {
                return;
            }
        }
        if (child_curval) {
            curval_str = val_make_sprintf_string(child_curval);
            if (curval_str == NULL) {
                if (newval_str) {
                    m__free(newval_str);
                }
                return;
            }
        }

        log_debug("\n        %s: editop=%s newval=%s curval=%s",
                  child_newval ? VAL_NAME(child_newval) : NCX_EL_NULL,
                  child_editop ? op_editop_name(child_editop) : NCX_EL_NONE,
                  child_newval ? newval_str : NCX_EL_NULL,
                  child_curval ? curval_str : NCX_EL_NULL);

        if (child_newval && val_set_by_default(child_newval)) {
            log_debug("\n        NEWVAL is default");
        }

        if (child_curval && val_set_by_default(child_curval)) {
            log_debug("\n        CURVAL is default");
        }

        m__free(newval_str);
        m__free(curval_str);

        child_edit = agt_cfg_next_child_edit(child_edit);
    }
} /* handle_merge */


/********************************************************************
* FUNCTION handle_other
*
* Dump values during any but not MERGE operation
*
********************************************************************/
static void handle_other (val_value_t *newval,
                          val_value_t *curval)
{
    print_children(newval, "newval");
    print_children(curval, "curval");

} /* handle_other */


/********************************************************************
* FUNCTION y_sethook_example_init_static_vars
*
* initialize module static variables
*
********************************************************************/
static void y_sethook_example_init_static_vars (void)
{
    sethook_example_mod = NULL;
    interfaces_obj = NULL;
    trigger_obj = NULL;
    interfaces_val = NULL;
    trigger_val = NULL;

    /* init your static variables here */

} /* y_sethook_example_init_static_vars */


/********************************************************************
* FUNCTION sethook_example_edit2
*
* Edit database object callback
*
* Path: container /interfaces/interface/state
* Path: list /interfaces/interface
* Path: container /interfaces
* Path: leaf /trigger
* Path: leaf /list-a
* Path: leaf /list-a/list-b
* Path: leaf /list-a/list-c
*
* Add object instrumentation in COMMIT phase.
*
* INPUTS:
*     see agt/agt_cb.h for details
*
* RETURNS:
*     error status
********************************************************************/
static status_t
    sethook_example_edit2 (ses_cb_t *scb,
                           rpc_msg_t *msg,
                           agt_cbtyp_t cbtyp,
                           op_editop_t editop,
                           val_value_t *newval,
                           val_value_t *curval)
{
    status_t res = NO_ERR;
    val_value_t *errorval = (curval) ? curval : newval;
    uint32 cur_errs = agt_get_error_count(msg);

    print_callback_info(errorval, cbtyp, editop, (const xmlChar *)"EDIT-2");

    switch (cbtyp) {
    case AGT_CB_VALIDATE:
        /* description-stmt validation here */
        break;
    case AGT_CB_APPLY:
        /* database manipulation done here */
        break;
    case AGT_CB_COMMIT:
        /* device instrumentation done here */
        switch (editop) {
        case OP_EDITOP_LOAD:
            break;
        case OP_EDITOP_MERGE:
            handle_merge(msg, newval, curval);
            break;
        case OP_EDITOP_REPLACE:
        case OP_EDITOP_CREATE:
        case OP_EDITOP_DELETE:
            handle_other(newval, curval);
            break;
        default:
            res = SET_ERROR(ERR_INTERNAL_VAL);
        }

        if (res == NO_ERR) {
            res = agt_check_cache(&trigger_val, newval, curval, editop);
        }

        break;
    case AGT_CB_ROLLBACK:
        /* undo device instrumentation here */
        break;
    default:
        res = SET_ERROR(ERR_INTERNAL_VAL);
    }

    if ((res != NO_ERR) && (cur_errs == agt_get_error_count(msg))) {
        agt_record_error(
            scb,
            &msg->mhdr,
            NCX_LAYER_CONTENT,
            res,
            NULL,
            (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
            errorval,
            (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
            errorval);
    }
    return res;

} /* sethook_example_edit2 */


/********************************************************************
* FUNCTION  sethook_callback
*
* Callback function for server object handler
* Used to provide a callback for a specific named object
*
* Set Hook:
*   trigger: delete /trigger
*   add_edit:
*       delete nodes: delete the whole container
*
*           path: /interfaces
*
*********************************************************************/
static status_t
    sethook_callback (ses_cb_t *scb,
                      rpc_msg_t *msg,
                      agt_cfg_transaction_t *txcb,
                      op_editop_t editop,
                      val_value_t  *newval,
                      val_value_t  *curval)
{
    log_debug("\nEnter SET Hook callback");

    status_t res = NO_ERR;
    val_value_t *errorval = (curval) ? curval : newval;

    const xmlChar *defpath =
        (const xmlChar *)"/sethook-ex:interfaces";

    switch (editop) {
    case OP_EDITOP_LOAD:
        break;
    case OP_EDITOP_MERGE:
    case OP_EDITOP_REPLACE:
    case OP_EDITOP_CREATE:
        break;
    case OP_EDITOP_DELETE:
        /* delete the interfaces container if the curval of 'trigger' node is
         * "delete-all"
         */
        if (curval &&
            !xml_strcmp(VAL_STR(curval), (const xmlChar *)"delete-all")) {

            res = agt_val_add_edit(scb,
                                   msg,
                                   txcb,
                                   defpath,
                                   NULL, // editval
                                   editop);
        }
        break;
    default:
        res = SET_ERROR(ERR_INTERNAL_VAL);
    }

    if (res != NO_ERR) {
        agt_record_error(scb,
                         &msg->mhdr,
                         NCX_LAYER_CONTENT,
                         res,
                         NULL,
                         (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
                         errorval,
                         (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
                         errorval);
    }

    return res;

}  /* sethook_callback */


/********************************************************************
* FUNCTION  sethook_callback2
*
* Callback function for server object handler
* Used to provide a callback for a specific named object
*
* Set Hook:
*   trigger: delete /list-a/list-c
*   add_edit:
*       delete nodes: delete the the list list-b entry
*
*           path: /list-a/list-b
*
*********************************************************************/
static status_t
    sethook_callback2 (ses_cb_t *scb,
                       rpc_msg_t *msg,
                       agt_cfg_transaction_t *txcb,
                       op_editop_t editop,
                       val_value_t *newval,
                       val_value_t *curval)
{
    log_debug("\nEnter SET Hook callback");

    status_t res = NO_ERR;
    val_value_t *errorval = (curval) ? curval : newval;
    val_value_t *key_val = NULL;

    const xmlChar *defpath =
        (const xmlChar *)"/list-a[name='name1']/list-b[name='name2']";

    /* find a list key value for verification */
    if (curval) {
        key_val =
            val_find_child(curval,
                           y_sethook_example_M_sethook_example,
                           (const xmlChar *)"name");

        if (key_val) {
            log_debug("\n>>>>>>>>>> "
                      " callback for %s editop, name=%s",
                       op_editop_name(editop), VAL_STR(key_val));
        }
    }

    switch (editop) {
    case OP_EDITOP_LOAD:
        break;
    case OP_EDITOP_MERGE:
    case OP_EDITOP_REPLACE:
    case OP_EDITOP_CREATE:
        break;
    case OP_EDITOP_DELETE:
        /* delete the list-b list if the curval list key value is
         * "delete-list-b"
         */
        if (key_val &&
            !xml_strcmp(VAL_STR(key_val), (const xmlChar *)"delete-list-b")) {

            res = agt_val_add_edit(scb,
                                   msg,
                                   txcb,
                                   defpath,
                                   NULL, // editval
                                   editop);
        }
        break;
    default:
        res = SET_ERROR(ERR_INTERNAL_VAL);
    }

    if (res != NO_ERR) {
        agt_record_error(scb,
                         &msg->mhdr,
                         NCX_LAYER_CONTENT,
                         res,
                         NULL,
                         (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
                         errorval,
                         (errorval) ? NCX_NT_VAL : NCX_NT_NONE,
                         errorval);
    }

    return res;

}  /* sethook_callback2 */


/********************************************************************
* FUNCTION y_sethook_example_init
*
* initialize the sethook-example server instrumentation library
*
* INPUTS:
*    modname == requested module name
*    revision == requested version (NULL for any)
*
* RETURNS:
*     error status
********************************************************************/
status_t
    y_sethook_example_init (const xmlChar *modname,
                            const xmlChar *revision)
{
    status_t res = NO_ERR;

    y_sethook_example_init_static_vars();

    /* change if custom handling done */
    if (xml_strcmp(modname, y_sethook_example_M_sethook_example)) {
        return ERR_NCX_UNKNOWN_MODULE;
    }

    if (revision && xml_strcmp(revision, y_sethook_example_R_sethook_example)) {
        return ERR_NCX_WRONG_VERSION;
    }

    res = ncxmod_load_module(y_sethook_example_M_sethook_example,
                             y_sethook_example_R_sethook_example,
                             agt_get_savedevQ(),
                             &sethook_example_mod);
    if (res != NO_ERR) {
        return res;
    }

    interfaces_obj =
        ncx_find_object(sethook_example_mod,
                        y_sethook_example_N_interfaces);
    if (interfaces_obj == NULL) {
        return ERR_NCX_DEF_NOT_FOUND;
    }

    trigger_obj =
        ncx_find_object(sethook_example_mod,
                        y_sethook_example_N_trigger);
    if (trigger_obj == NULL) {
        return ERR_NCX_DEF_NOT_FOUND;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                                       (const xmlChar *)"/sethook-ex:interfaces",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                (const xmlChar *)"/sethook-ex:interfaces/sethook-ex:interface",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
            (const xmlChar *)"/sethook-ex:interfaces/sethook-ex:interface/sethook-ex:state",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_callback(y_sethook_example_M_sethook_example,
                                 (const xmlChar *)"/sethook-ex:trigger",
                                 y_sethook_example_R_sethook_example,
                                 sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                                       (const xmlChar *)"/sethook-ex:list-a",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                        (const xmlChar *)"/sethook-ex:list-a/sethook-ex:list-b",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }

    res =
        agt_cb_register_edit2_callback(y_sethook_example_M_sethook_example,
                        (const xmlChar *)"/sethook-ex:list-a/sethook-ex:list-c",
                                       y_sethook_example_R_sethook_example,
                                       sethook_example_edit2);
    if (res != NO_ERR) {
        return res;
    }


    /* put your module initialization code here */


    /* Set Hook callback registration */
    res = agt_cb_hook_register((const xmlChar *)"/sethook-ex:trigger",
                               AGT_HOOKFMT_NODE,
                               AGT_HOOK_TYPE_SETHOOK,
                               sethook_callback);
    if (res != NO_ERR) {
        return res;
    }

    res = agt_cb_hook_register((const xmlChar *)"/sethook-ex:list-a/sethook-ex:list-c",
                               AGT_HOOKFMT_NODE,
                               AGT_HOOK_TYPE_SETHOOK,
                               sethook_callback2);
    if (res != NO_ERR) {
        return res;
    }

    return res;

} /* y_sethook_example_init */


/********************************************************************
* FUNCTION y_sethook_example_init2
*
* SIL init phase 2: non-config data structures
* Called after running config is loaded
*
* RETURNS:
*     error status
********************************************************************/
status_t y_sethook_example_init2 (void)
{
    status_t res = NO_ERR;

    interfaces_val =
        agt_init_cache(y_sethook_example_M_sethook_example,
                       y_sethook_example_N_interfaces,
                       &res);
    if (res != NO_ERR) {
        return res;
    }

    trigger_val =
        agt_init_cache(y_sethook_example_M_sethook_example,
                       y_sethook_example_N_trigger,
                       &res);
    if (res != NO_ERR) {
        return res;
    }

    /* put your init2 code here */

    return res;

} /* y_sethook_example_init2 */


/********************************************************************
* FUNCTION y_sethook_example_cleanup
*    cleanup the server instrumentation library
*
********************************************************************/
void y_sethook_example_cleanup (void)
{

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:interfaces");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:interfaces/sethook-ex:interface");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:interfaces/sethook-ex:interface/sethook-ex:state");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:trigger");


    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:list-a");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:list-a/sethook-ex:list-b");

    agt_cb_unregister_callbacks(
        y_sethook_example_M_sethook_example,
        (const xmlChar *)"/sethook-ex:list-a/sethook-ex:list-c");


    /* put your cleanup code here */


    /* Set Hook callback clean up */
    agt_cb_hook_unregister((const xmlChar *)"/sethook-ex:trigger");

    agt_cb_hook_unregister((const xmlChar *)"/sethook-ex:list-a/sethook-ex:list-c");


} /* y_sethook_example_cleanup */

/* END SIL sethook_example.c */
